/*
 * This file is part of the Disco Deterministic Network Calculator v2.2.0 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2008 - 2010 Andreas Kiefer
 * Copyright (C) 2011 - 2015 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.HashMap;
import java.util.Map.Entry;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.nc.AnalysisConfig.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.network.Server.Multiplexing;
import unikl.disco.numbers.Num;
import unikl.disco.minplus.Convolution;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 * @author Steffen Bondorf
 * 
 */
public class SeparateFlowAnalysis extends Analysis {
	@SuppressWarnings("unused")
	private SeparateFlowAnalysis() {}

	public SeparateFlowAnalysis( Network network ) {
		super( network );
		super.result = new SeparateFlowAnalysisResults();
	}

	public SeparateFlowAnalysis( Network network, AnalysisConfig configuration ) {
		super( network, configuration );
		super.result = new SeparateFlowAnalysisResults();
	}
	
	/**
	 * Performs a separated flow analysis for the <code>flow_of_interest</code>.
	 * 
	 * This analysis first blends out the flow of interest and then computes for
	 * each server along this flow's path the left-over service curve that
	 * results if all remaining flows crossing this server receive their maximum
	 * amount of service. Then all left-over service curves are concatenated to
	 * receive the end-to-end service curve from the perspective of the flow of
	 * interest.
	 * 
	 */
	public void performAnalysis( Flow flow_of_interest ) throws Exception {
		performAnalysis( flow_of_interest, flow_of_interest.getPath() );
	}

	public void performAnalysis( Flow flow_of_interest, Path path ) throws Exception {
		((SeparateFlowAnalysisResults) result).betas_e2e = getServiceCurves( flow_of_interest, path, Collections.singleton( flow_of_interest ) );

		Num delay_bound__beta_e2e;
		Num backlog_bound__beta_e2e;
		
		result.delay_bound = Num.createPositiveInfinity();
		result.backlog_bound = Num.createPositiveInfinity();
		
		for( ServiceCurve beta_e2e : ((SeparateFlowAnalysisResults) result).betas_e2e ) {
			delay_bound__beta_e2e = DelayBound.deriveFIFO( flow_of_interest.getArrivalCurve(), beta_e2e ); // single flow of interest, i.e., fifo per micro flow holds
			if( delay_bound__beta_e2e.le( result.delay_bound ) ) {
				result.delay_bound = delay_bound__beta_e2e;
			}
			
			backlog_bound__beta_e2e = BacklogBound.derive( flow_of_interest.getArrivalCurve(), beta_e2e );
			if( backlog_bound__beta_e2e.le( result.backlog_bound ) ) {
				result.backlog_bound = backlog_bound__beta_e2e;
			}
		}
		
		result.succeeded = true;
	}

	protected Set<ServiceCurve> getServiceCurves( Flow flow_of_interest, Path path, Set<Flow> flows_to_serve ) throws Exception {
		Set<ServiceCurve> betas_lofoi_path = new HashSet<ServiceCurve>();
		Set<ServiceCurve> betas_lofoi_s;
		
		// Convolve all left over service curves, server by server
		for ( Server server : path.getServers() ) {
			betas_lofoi_s = new HashSet<ServiceCurve>();
			
			Set<Flow> f_xfoi_server = network.getFlows( server );
			f_xfoi_server.remove( flow_of_interest );
			f_xfoi_server.removeAll( flows_to_serve );

			ServiceCurve beta_lofoi = server.getServiceCurve();
			
			if ( f_xfoi_server.isEmpty() ) {
				betas_lofoi_s.add( beta_lofoi );
			} else {															// network, server, flows_to_bound, flows_lower_priority
				Set<ArrivalCurve> alpha_xfois = ArrivalBound.computeArrivalBounds( network, configuration, server, f_xfoi_server, flow_of_interest );
				
				// Calculate the left-over service curve for the flow of interest
				if( configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO
					|| ( configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL && server.multiplexingDiscipline() == Multiplexing.FIFO ) )
				{
					for ( ArrivalCurve alpha_xfoi : alpha_xfois ) {
						betas_lofoi_s.add( LeftOverService.fifoMux( beta_lofoi, alpha_xfoi ) );
					}
				}
				else
				{
					for ( ArrivalCurve alpha_xfoi : alpha_xfois ) {
						betas_lofoi_s.add( LeftOverService.arbMux( beta_lofoi, alpha_xfoi ) );
					}
				}
				result.map__server__alphas.put( server, alpha_xfois );
			}
			((SeparateFlowAnalysisResults) result).map__server__betas_lo.put( server, betas_lofoi_s );
			
			betas_lofoi_path = Convolution.convolve_SCs_SCs( betas_lofoi_path, betas_lofoi_s, configuration.tbrlConvolution() );
		}
		return betas_lofoi_path;
	}

	public Set<ServiceCurve> getLeftOverServiceCurves() {
		return ((SeparateFlowAnalysisResults) result).betas_e2e;
	}
	
	public Map<Server,Set<ServiceCurve>> getServerLeftOverBetasMap(){
		return ((SeparateFlowAnalysisResults) result).map__server__betas_lo;
	}
	
	public String getServerLeftOverBetasMapString() {
		return ((SeparateFlowAnalysisResults) result).getServerLeftOverBetasMapString();
	}
}

class SeparateFlowAnalysisResults extends AnalysisResults {
	protected Set<ServiceCurve> betas_e2e;
	protected Map<Server,Set<ServiceCurve>> map__server__betas_lo;

	protected SeparateFlowAnalysisResults(){
		super();
		betas_e2e = new HashSet<ServiceCurve>();
		map__server__betas_lo = new HashMap<Server,Set<ServiceCurve>>();
	}
	
	protected SeparateFlowAnalysisResults( Num delay_bound,
						 Num backlog_bound,
						 Set<ServiceCurve> betas_e2e,
						 Map<Server,Set<ServiceCurve>> map__server__betas_lo,
						 Map<Server,Set<ArrivalCurve>> map__server__alphas ) {
		
		super( true, delay_bound, backlog_bound, map__server__alphas );
		
		this.betas_e2e = betas_e2e;
		this.map__server__betas_lo = map__server__betas_lo;
	}

	public String getServerLeftOverBetasMapString() {
		if( !succeeded ) {
			return "Analysis failed";
		}
		
		if( map__server__betas_lo.isEmpty() ) {
			return "{}";
		}

		StringBuffer result_str = new StringBuffer( "{" );

		for( Entry<Server,Set<ServiceCurve>> entry : map__server__betas_lo.entrySet() ) {
			result_str.append( entry.getKey().toShortString() );
			result_str.append( "={" );
			for ( ServiceCurve beta_lo : entry.getValue() ) {
				result_str.append( beta_lo.toString() );
				result_str.append(  "," );
			}
			result_str.deleteCharAt( result_str.length()-1 ); // Remove the trailing comma.
			result_str.append( "}" );
			result_str.append( ", " );
		}
		result_str.delete( result_str.length()-2, result_str.length() ); // Remove the trailing blank space and comma.
		
		result_str.append( "}" );
		
		return result_str.toString();
	}
}